@@ -1,4 +1,4 @@
-# Copyright (C) 2007 AG Projects. See LICENSE for details.
+# Copyright (C) 2007-2008 AG Projects. See LICENSE for details.
 #
 
 """GNUTLS connection support"""
@@ -8,6 +8,7 @@ __all__ = ['X509Credentials', 'ClientSession', 'ServerSession', 'ServerSessionFa
 from time import time
 from socket import SHUT_RDWR as SOCKET_SHUT_RDWR
 
+from _ctypes import PyObj_FromPtr
 from ctypes import *
 
 from gnutls.validators import *
@@ -21,10 +22,52 @@ from gnutls.library.constants import GNUTLS_CERT_SIGNER_NOT_FOUND, GNUTLS_CERT_S
 from gnutls.library.constants import GNUTLS_AL_FATAL, GNUTLS_A_BAD_CERTIFICATE
 from gnutls.library.constants import GNUTLS_A_UNKNOWN_CA, GNUTLS_A_INSUFFICIENT_SECURITY
 from gnutls.library.constants import GNUTLS_A_CERTIFICATE_EXPIRED, GNUTLS_A_CERTIFICATE_REVOKED
+from gnutls.library.constants import GNUTLS_NAME_DNS
 from gnutls.library.types     import gnutls_certificate_credentials_t, gnutls_session_t, gnutls_x509_crt_t
+from gnutls.library.types     import gnutls_certificate_server_retrieve_function
 from gnutls.library.functions import *
 
 
+@gnutls_certificate_server_retrieve_function
+def _retrieve_server_certificate(c_session, retr_st):
+    session = PyObj_FromPtr(gnutls_session_get_ptr(c_session))
+    identity = session.credentials.select_server_identity(session)
+    retr_st.contents.type = GNUTLS_CRT_X509
+    retr_st.contents.deinit_all = 0
+    if identity is None:
+        retr_st.contents.ncerts = 0
+    else:
+        retr_st.contents.ncerts = 1
+        retr_st.contents.cert.x509.contents = identity.cert._c_object
+        retr_st.contents.key.x509 = identity.key._c_object
+    return 0
+
+
+class _ServerNameIdentities(dict):
+    """Used internally by X509Credentials to map server names to X509 identities for the server name extension"""
+    def __init__(self, identities):
+        dict.__init__(self)
+        for identity in identities:
+            self.add(identity)
+    def add(self, identity):
+        for name in identity.cert.alternative_names.dns:
+            self[name.lower()] = identity
+        for ip in identity.cert.alternative_names.ip:
+            self[ip] = identity
+        subject = identity.cert.subject
+        if subject.CN is not None:
+            self[subject.CN.lower()] = identity
+    def get(self, server_name, default=None):
+        server_name = server_name.lower()
+        if server_name in self:
+            return self[server_name]
+        for name in (n for n in self if n.startswith('*.')):
+            suffix = name[1:]
+            if server_name.endswith(suffix) and '.' not in server_name[:-len(suffix)]:
+                return self[name]
+        return default
+
+
 class X509Credentials(object):
     DH_BITS  = 1024
     RSA_BITS = 1024
@@ -40,28 +83,34 @@ class X509Credentials(object):
         instance._c_object = c_object
         return instance
 
-    @method_args((X509Certificate, none), (X509PrivateKey, none), list_of(X509Certificate), list_of(X509CRL))
-    def __init__(self, cert=None, key=None, trusted=[], crl_list=[]):
-        """Credentials object containing an X509 certificate, a private key and 
-           optionally a list of trusted CAs and a list of CRLs."""
+    @method_args((X509Certificate, none), (X509PrivateKey, none), list_of(X509Certificate), list_of(X509CRL), list_of(X509Identity))
+    def __init__(self, cert=None, key=None, trusted=[], crl_list=[], identities=[]):
+        """Credentials contain a X509 certificate, a private key, a list of trusted CAs and a list of CRLs (all optional).
+        An optional list of additional X509 identities can be specified for applications that need more that one identity"""
         if cert and key:
             gnutls_certificate_set_x509_key(self._c_object, byref(cert._c_object), 1, key._c_object)
         elif (cert, key) != (None, None):
             raise ValueError("Specify neither or both the certificate and private key")
-        # this generates core dumping - gnutls_certificate_set_params_function(self._c_object, gnutls_params_function(self.__get_params))
+        gnutls_certificate_server_set_retrieve_function(self._c_object, _retrieve_server_certificate)
         self._max_depth = 5
         self._max_bits  = 8200
         self._type = CRED_CERTIFICATE
+        self._cert = cert
+        self._key = key
+        self._identities = tuple(identities)
         self._trusted = ()
-        self.cert = cert
-        self.key = key
         self.add_trusted(trusted)
         self.crl_list = crl_list
+        self.server_name_identities = _ServerNameIdentities(identities)
+        if cert and key:
+            self.server_name_identities.add(X509Identity(cert, key))
         self.session_params = SessionParams(self._type)
 
     def __del__(self):
         self.__deinit(self._c_object)
 
+    # Methods to alter the credentials at runtime
+
     @method_args(list_of(X509Certificate))
     def add_trusted(self, trusted):
         size = len(trusted)
@@ -80,15 +129,20 @@ class X509Credentials(object):
         X509Credentials.rsa_params = RSAParams(bits)
         del reference
 
-    def __get_params(self, session, type, st):
-        """Callback function that is used when a session requests DH or RSA parameters"""
-        # static int get_params( gnutls_session_t session, gnutls_params_type_t type, gnutls_params_st *st)
-        # see example http://www.gnu.org/software/gnutls/manual/gnutls.html#Parameters-stored-in-credentials -Mircea
-        print "get_params callback:", session, type, st
-        return 0
-
     # Properties
 
+    @property
+    def cert(self):
+        return self._cert
+
+    @property
+    def key(self):
+        return self._key
+
+    @property
+    def identities(self):
+        return self._identities
+
     @property
     def trusted(self):
         return self._trusted
@@ -119,6 +173,8 @@ class X509Credentials(object):
     max_verify_bits = property(_get_max_verify_bits, _set_max_verify_bits)
     del _get_max_verify_bits, _set_max_verify_bits
 
+    # Methods to select and validate certificates
+
     def check_certificate(self, cert, cert_name='certificate'):
         """Verify activation, expiration and revocation for the given certificate"""
         now = time()
@@ -129,6 +185,17 @@ class X509Credentials(object):
         for crl in self.crl_list:
             crl.check_revocation(cert, cert_name=cert_name)
 
+    def select_server_identity(self, session):
+        """Select which identity the server will use for a given session. The default selection algorithm uses
+        the server name extension. A subclass can overwrite it if a different selection algorithm is desired."""
+        server_name = session.server_name
+        if server_name is not None:
+            return self.server_name_identities.get(server_name)
+        elif self.cert and self.key:
+            return self ## since we have the cert and key attributes we can behave like a X509Identity
+        else:
+            return None
+
 
 class SessionParams(object):
     _default_kx_algorithms = {
@@ -208,6 +275,8 @@ class Session(object):
 
     def __init__(self, socket, credentials):
         gnutls_init(byref(self._c_object), self.session_type)
+        ## Store a pointer to self on the C session
+        gnutls_session_set_ptr(self._c_object, id(self))
         # gnutls_dh_set_prime_bits(session, DH_BITS)?
         gnutls_transport_set_ptr(self._c_object, socket.fileno())
         gnutls_handshake_set_private_extensions(self._c_object, 1)
@@ -356,6 +425,21 @@ class Session(object):
 class ClientSession(Session):
     session_type = GNUTLS_CLIENT
 
+    def __init__(self, socket, credentials, server_name=None):
+        Session.__init__(self, socket, credentials)
+        self._server_name = None
+        if server_name is not None:
+            self.server_name = server_name
+
+    def _get_server_name(self):
+        return self._server_name
+    @method_args(str)
+    def _set_server_name(self, server_name):
+        gnutls_server_name_set(self._c_object, GNUTLS_NAME_DNS, c_char_p(server_name), len(server_name))
+        self._server_name = server_name
+    server_name = property(_get_server_name, _set_server_name)
+    del _get_server_name, _set_server_name
+
 
 class ServerSession(Session):
     session_type = GNUTLS_SERVER
@@ -364,6 +448,25 @@ class ServerSession(Session):
         Session.__init__(self, socket, credentials)
         gnutls_certificate_server_set_request(self._c_object, CERT_REQUEST)
 
+    @property
+    def server_name(self):
+        data_length = c_size_t(256)
+        data = create_string_buffer(data_length.value)
+        hostname_type = c_uint()
+        for i in xrange(2**16):
+            try:
+                gnutls_server_name_get(self._c_object, data, byref(data_length), byref(hostname_type), i)
+            except RequestedDataNotAvailable:
+                break
+            except MemoryError:
+                data_length.value += 1 ## one extra byte for the terminating 0
+                data = create_string_buffer(data_length.value)
+                gnutls_server_name_get(self._c_object, data, byref(data_length), byref(hostname_type), i)
+            if hostname_type.value != GNUTLS_NAME_DNS:
+                continue
+            return data.value
+        return None
+
 
 class ServerSessionFactory(object):
 
